#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/ioctl.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/uaccess.h>

#include "fpga_io.h"

static char io_data[LEN];
static unsigned int busy = 0;

/*システムコールopen(...)によって呼ばれる*/
static int fpga_io_open(struct inode *inode, struct file *filp){
	int err;

	if(busy) return -EBUSY; //一つしか開けない。マイナスにすることで、エラーを表す。
	else busy = 1;
	//printk(" filp->f_version : %lu \n",filp->f_version); //f_version はopenの度に異なる値をもつ。
	//MOD_INC_USE_COUNT; //0より大きいとモジュールがはずせなくなる。

	if((err = check_region(FPGA_IO_PORT_READ1,1)) < 0 ){
		printk(" This ports <%x - %x> are using now!\n",FPGA_IO_PORT_READ1,1);
		printk(" check_region(port,range) : %d\n",err);
		return err;
	}
	if((err = check_region(FPGA_IO_PORT_READ2,1)) < 0 ){
		printk(" This ports <%x - %x> are using now!\n",FPGA_IO_PORT_READ2,1);
		printk(" check_region(port,range) : %d\n",err);
		return err;
	}
	
	request_region(FPGA_IO_PORT_READ1,1,"fpga_io");
	request_region(FPGA_IO_PORT_READ2,1,"fpga_io");
	return 0;
}

/*システムコールread(...)によって呼ばれる*/
/*offset も count も使用していない．*/
static ssize_t fpga_io_read(struct file *file, char *buf, size_t count, loff_t *offset){
	int data1=0;
	//int data2=0;
	int i=0;

	data1 = inb(FPGA_IO_PORT_READ1) & 0x00ff;
	//data2 = inb(FPGA_IO_PORT_READ2) & 0x000f;

	//printk("data1 = %x\ndata2 = %x\n",data_1,data_2);
	//for(i=3;i>=0;i--){
	//	if(data2 >= (0x0001 << i)){ io_data[i+8] = '1'; data2 -= (0x0001 << i); }
	//	else{ io_data[i+8] = '0';}
	//}
	for(i=7;i>=0;i--){
		if(data1 >= (0x0001 << i)){ io_data[i] = '1'; data1 -= (0x0001 << i); }
		else{ io_data[i] = '0'; }
	}
 
	copy_to_user(buf,io_data,LEN); //io_dataをユーザー空間にコピーする
	printk(" requested : %d bytes returned: %d bytes\n",count,LEN);
	return LEN;

}

static int fpga_io_write(struct file *file, const char *buff, size_t count, loff_t * offset){
	int i,len = 4;
	unsigned char out_data = 0;
	printk("count : %d\n",count);
	if(len > count) len=count;
	copy_from_user(io_data,buff,len);
	for(i=0;i<len;i++){
		if(io_data[i]=='1') out_data += (0x01 << i);
	}
	printk("writing : %d\n", out_data);
	outb(out_data,FPGA_IO_PORT_READ2);
	return len;	
}

/*システムコールclose(...)によって呼ばれる*/
static int fpga_io_release(struct inode *inode, struct file *filp){
	busy = 0;
	//MOD_DEC_USE_COUNT; //0より大きいとモジュールがはずせなくなる。

	//printk(" filp->f_version : %lu \n",filp->f_version); //f_version はopenの度に異なる値をもつ。

	//release_region(FPGA_IO_PORT_READ1,FPGA_IO_PORT_MASK);
	release_region(FPGA_IO_PORT_READ1,1);
	release_region(FPGA_IO_PORT_READ2,1);
	//printk(" I/Oport from: %x to: %x \n",FPGA_IO_PORT_READ1,1);
	return 0;
}

//カーネルへ関数を登録。システムコールと関数の関連付け。
static struct file_operations fpga_io_fops = {
	owner:							THIS_MODULE,
	read:							fpga_io_read,
	write:							fpga_io_write,
	ioctl:							NULL,
	open:							fpga_io_open,
	release:						fpga_io_release,
};

int init_module (void){
	printk("install 'fpga_io' into major %d\n",FPGA_IO_MAJOR);
	if(register_chrdev(FPGA_IO_MAJOR, "fpga_io", &fpga_io_fops)){ /*登録*/
		printk("device registration error\n");
		return -EBUSY;
	}
	return 0;
}

void cleanup_module (void){
	busy = 0;
	printk("remove 'fpga_io' from major %d\n",FPGA_IO_MAJOR);
	unregister_chrdev(FPGA_IO_MAJOR,"fpga_io"); /*登録解除*/
}
